"""rio-tiler.mosaic.methods: Mosaic filling methods."""

from .defaults import (  # noqa
    FirstMethod,
    HighestMethod,
    LastBandHigh,
    LastBandLow,
    LowestMethod,
    MeanMethod,
    MedianMethod,
    StdevMethod,
)
